package refit.replica.order;

import refit.config.REFITConfig;
import refit.util.REFITAssert;

public class REFITOrderGroups {

	public static final int COUNT = REFITConfig.ORDER_GROUPS.length;
	private static final boolean[][] orderGroupReplicas;

	static {
		if (REFITConfig.MAXIMUM_BATCH_SIZE.length != COUNT) {
			throw new AssertionError("MAXIMUM_BATCH_SIZE="
					+ REFITConfig.MAXIMUM_BATCH_SIZE.length + " must have the same length as ORDER_GROUPS=" + COUNT + "!");
		}
		if (REFITConfig.ORDER_CLIENT_WEIGHTS.length != COUNT) {
			throw new AssertionError("ORDER_CLIENT_WEIGHTS="
					+ REFITConfig.ORDER_CLIENT_WEIGHTS.length + " must have the same length as ORDER_GROUPS=" + COUNT + "!");
		}
		for (int i = 0; i < COUNT; i++) {
			if (REFITConfig.ORDER_GROUPS[i].length != REFITConfig.ORDER_GROUP_SIZE) {
				throw new AssertionError("The length of ORDER_GROUPS[" + i + "] must be ORDER_GROUP_SIZE!");
			}
		}

		orderGroupReplicas = new boolean[REFITConfig.ORDER_GROUPS.length][];
		for (int i = 0; i < orderGroupReplicas.length; i++) {
			orderGroupReplicas[i] = new boolean[REFITConfig.TOTAL_NR_OF_REPLICAS];
			for (short replicaID : REFITConfig.ORDER_GROUPS[i]) {
				orderGroupReplicas[i][replicaID] = true;
			}
		}
	}

	public static short getGroupID(long instanceID) {
		return (short) (instanceID % COUNT);
	}

	public static boolean[] getGroup(long instanceID) {
		return getGroupByID(getGroupID(instanceID));
	}

	public static boolean isInGroup(short replicaID, long instanceID) {
		return getGroup(instanceID)[replicaID];
	}

	public static boolean[] getGroupByID(int groupID) {
		return orderGroupReplicas[groupID];
	}

	public static boolean isInGroupByID(short replicaID, int groupID) {
		return getGroupByID(groupID)[replicaID];
	}


	public static short getPrimary(long instanceID, int viewID) {
		return getGroupPrimary(getGroupID(instanceID), viewID);
	}

	public static short getGroupPrimary(int groupID, int viewID) {
		short[] orderingGroup = REFITConfig.ORDER_GROUPS[groupID];
		return orderingGroup[viewID % REFITConfig.ORDER_GROUP_LEADER_LIMIT[groupID]];
	}

	public static long nextSequenceNumber(long instanceID, int groupID) {
		do {
			instanceID++;
		} while (getGroupID(instanceID) != groupID);
		return instanceID;
	}

	public static short getPrimaryGroup(short replicaId) {
		for (short i = 0; i < REFITConfig.ORDER_GROUPS.length; i++) {
			for (int j = 0; j < REFITConfig.ORDER_GROUP_LEADER_LIMIT[i]; j++) {
				if (REFITConfig.ORDER_GROUPS[i][j] == replicaId)
					return i;
			}
		}

		REFITAssert.fail("Replica " + replicaId + " is not in any replica group");
		return -1;
	}

	public static short getClientGroup(int id) {
		if (REFITConfig.CLIENT_FORCE_GROUP >= 0) {
			return REFITConfig.CLIENT_FORCE_GROUP;
		}
		for (short i = 0; i < REFITConfig.TOTAL_NR_OF_REPLICAS; i++) {
			if (REFITConfig.LOCATION[id] == REFITConfig.LOCATION[i]) {
				return getPrimaryGroup(i);
			}
		}
		throw new IllegalArgumentException("No order group for client");
	}
}
